/*
    ShapeSpatialPakTemplateCmd.h - function plotter for spatial templates

    Will Eighteen, 2001: will@oplaine.demon.co.uk
*/

#ifndef SHAPESPATIALPAKTEMPLATECMD_H
#define SHAPESPATIALPAKTEMPLATECMD_H

#include <NoUndoCmd.h>

#ifndef BREAK
#define BREAK
#include <BreakCmd.h>
#endif


class GraphIntCmd;
class GraphFltCmd;
class MixwaveUtils;
class DataConformCmd;


class ShapeSpatialPakTemplateCmd : public NoUndoCmd
{
	public:
		ShapeSpatialPakTemplateCmd ( char *, int, unsigned int, unsigned int );
		~ShapeSpatialPakTemplateCmd();
		void updateSelection ( unsigned int, long, long, long, long, long, int *, float * );
		void templateID ( char *, long, long, long, long, long, float, float,
							GraphIntCmd *, GraphFltCmd *, int *, float *, int,
							unsigned int );
		unsigned int dataChanged ( void ) { return _dataChanged; }
		void reset() { _dataChanged = 0; }

	protected:
		virtual void doCommand();	// ShapeSpatialPakTemplate

	private:
		static void dataConformFinishedCallback ( BreakCmd *, Boolean, void * );
		void dataConformFinished ( DataConformCmd * );

		// callback for WM 'X'
		static void clientMessageProc ( Widget, XtPointer, XtPointer );

		// function button callbacks
		static void funct1Callback ( Widget, XtPointer, XtPointer );
		static void funct2Callback ( Widget, XtPointer, XtPointer );
		static void funct3Callback ( Widget, XtPointer, XtPointer );
		static void funct4Callback ( Widget, XtPointer, XtPointer );
		static void funct5Callback ( Widget, XtPointer, XtPointer );
		static void funct6Callback ( Widget, XtPointer, XtPointer );
		static void funct7Callback ( Widget, XtPointer, XtPointer );

		// Callbacks for data change by user
		static void field1aEntryCallback ( Widget, XtPointer, XtPointer );
		static void field1bEntryCallback ( Widget, XtPointer, XtPointer );

		static void field2aEntryCallback ( Widget, XtPointer, XtPointer );
		static void field2bEntryCallback ( Widget, XtPointer, XtPointer );

		static void field3aEntryCallback ( Widget, XtPointer, XtPointer );
		static void field3bEntryCallback ( Widget, XtPointer, XtPointer );

		static void field4aEntryCallback ( Widget, XtPointer, XtPointer );
		static void field4bEntryCallback ( Widget, XtPointer, XtPointer );

		static void field5aEntryCallback ( Widget, XtPointer, XtPointer );
		static void field5bEntryCallback ( Widget, XtPointer, XtPointer );

		static void field6aEntryCallback ( Widget, XtPointer, XtPointer );
		static void field6bEntryCallback ( Widget, XtPointer, XtPointer );

		static void field7aEntryCallback ( Widget, XtPointer, XtPointer );
		static void field7bEntryCallback ( Widget, XtPointer, XtPointer );

		// fns called when data in text box is changed
		void field1aEntry ( char * );
		void field1bEntry ( char * );
		void changeField1a ( float );
		void changeField1b ( float );

		void field2aEntry ( char * );
		void field2bEntry ( char * );
		void changeField2a ( float );
		void changeField2b ( float );

		void field3aEntry ( char * );
		void field3bEntry ( char * );
		void changeField3a ( float );
		void changeField3b ( float );

		void field4aEntry ( char * );
		void field4bEntry ( char * );
		void changeField4a ( float );
		void changeField4b ( float );

		void field5aEntry ( char * );
		void field5bEntry ( char * );
		void changeField5a ( float );
		void changeField5b ( float );

		void field6aEntry ( char * );
		void field6bEntry ( char * );
		void changeField6a ( float );
		void changeField6b ( float );

		void field7aEntry ( char * );
		void changeField7a ( unsigned int );
		void field7bEntry ( char * );
		void changeField7b ( float );

		void setTitle ( char * );
		int checkTextEntry ( char * );
		int readData();
		void writeData();
		void funct1 ( float, float );	// line interpolation
		void funct2 ( float, float );	// sinusoid
		void funct3 ( float, float );	// exponential
		void funct4 ( float, float );	// natural log.
		void funct5 ( float, float );	// natural log.
		void funct6 ( float, float );	// shift/scale
		void funct7 ( unsigned int, float );	// curve fit

		GraphIntCmd *_gic;
		GraphFltCmd *_gfc;
		MixwaveUtils *_utils;
		Boolean _shapeTemplateRealised;
		char _fileName[1024];
		int *_idata;
		float *_fdata;
		const char *_name;
		long _startData, _endData, _dataOffset;
		int _dataType;
		unsigned int _numDigits;
		unsigned int _fpDigits;
		unsigned int _dataChanged;
		unsigned int _viewID;	// cached owner of current selection

		long _startRange;
		long _endRange;
		long _startSel;
		long _endSel;
		long _xCurrent;
		float _min;
		float _max;

		Widget _rc1, _funct1, _label1a, _label1b, _field1a, _field1b;
		Widget _rc2, _funct2, _label2a, _label2b, _field2a, _field2b;
		Widget _rc3, _funct3, _label3a, _label3b, _field3a, _field3b;
		Widget _rc4, _funct4, _label4a, _label4b, _field4a, _field4b;
		Widget _rc5, _funct5, _label5a, _label5b, _field5a, _field5b;
		Widget _rc6, _funct6, _label6a, _label6b, _field6a, _field6b;
		Widget _rc7, _funct7, _label7a, _field7a, _label7b, _field7b;
		Widget _popup, _form, _label, _sep, _quit;
};


#endif

/*
   Notes:
   This class is due to a design failure. Spatial data is maintained in a
   dedicated pak, whereas for consistency it ought to be treated as a
   collection of individual templates. Therefore the classes which modify
   indivIdual pak data can't be used here. The whole spatial subsystem needs a
   rewrite to use individual template pak structures.

   Warning: use of small, >1, numbers in field7 result in fileLength/number
   mem. usage.
*/
