// v. 0.31 21/11/2001 w 18: added misc. header struct functions
// v. 0.3 01/01/2001 Will Eighteen: added HeaderSize, revised to default
// load r/w, fail to r.
// Waveclass v. 0.2: since it now does something useful! 19/04/1998
// Wave File Class: v. 0.1: last revision: 27/06/1998 by Will Eighteen

#ifndef X_WAVEFILE_CPP
#define X_WAVEFILE_CPP

//  ## (  Constant Definitions  ) ##
#define WAVE_FORMAT_PCM 1

//  ## (  Compiler 'C' Includes  ) ##
#include <math.h>
#include <iostream>

//  ## (  Custom Includes  ) ##
#include "waveclass.h"

//  ## (  Macro Definitions  ) ##
#define CHECKTYPE(a,b) ((a[0] == b[0]) && (a[1] == b[1]) && (a[2] == b[2]) && (a[3] == b[3]))


Wave *Wave::pFirst = 0;	// allocate space for statics
unsigned int Wave::_waveCount = 0;

// WAVECLASS FUNCTIONS

// private function: internal use only
Boolean Wave::IsWavFile(void)
{  //ok, but doesn't return wav type endianness
   char waveIdMinor[4] = {'W','A','V','E'};
   
   if(CHECKTYPE(waveIdMinor,waveHeader.header.rifftype))
      return True;	// PC's running DOS / MS Windows wave editors
               
   return False;		// not an identifiable wave file
}

Boolean Wave::IsAltWavFile(void)
{  //ok, but doesn't retun wav type endianness
   char waveIdMajor[4] = {'E','V','A','W'};
    
   if(CHECKTYPE(waveIdMajor,waveHeader.header.rifftype))
      return True;	// PC's running X Windows editors
                  
   return False;		// not an identifiable wave file
}

// Constructor: Ensure potential wave exists

Wave::Wave()
{
   wavFile = 0;
   pData = 0;
   pName = 0;
   // add self to beginning of list
   pNext = pFirst;
   pFirst = this;
   Init();
   _waveCount++;
   _waveNumber = _waveCount;
}


// Destructor: Cleanup on exit
Wave::~Wave()
{     
   // remove self from list
   if (pFirst == this)	// if I'm first..
   {
      pFirst = pNext; 	// .. then list is what's subsequent
   }
   else	// look for me
   {
      for (Wave *pW = pFirst; pW; pW = pW->pNext)
      {
      	if (pW->pNext == this)	// once found..
         {
            pW->pNext = pNext; 	// removed.
            break;
         }
      }
   }

   if(wavFile)
      fclose(wavFile);   

   Free();   
	_waveNumber = 0;
   _waveCount--;
}


// Deallocate wav data
void Wave::Free(void)
{
   if(pName)
      delete pName;
   
   if(pData)
      delete pData;
}


// Declared as private: internal use only
void Wave::Init(void)
{  
   memcpy(waveHeader.header.riff , (const void *)"RIFF", 4);
   waveHeader.header.filesize = 44;
   memcpy(waveHeader.header.rifftype, (const void *)"WAVE", 4); 
   waveHeader.ch_format.chunksize = 16; // dubious: should be supplied...
   memcpy(waveHeader.ch_format.chunk_id, (const void *)"fmt ", 4);
   waveHeader.format.wf.wFormatTag = WAVE_FORMAT_PCM;
   memcpy(waveHeader.ch_data.chunk_id, (const void *)"data", 4);
   waveHeader.ch_data.chunksize = 0;
}


// Declared as private: internal use only
void Wave::Set(short channels,long sampleRate,short dataSize)
{
   waveHeader.format.wf.nChannels=channels;
   waveHeader.format.wf.nSamplesPerSec=sampleRate;
   waveHeader.format.wf.nAvgBytesPerSec=sampleRate * (dataSize*channels/8);
   waveHeader.format.wf.nBlockAlign=dataSize*channels/8;
   waveHeader.format.wBitsPerSample=dataSize;   
}

// public functions

unsigned int Wave::waveNum(void)
{
	return _waveNumber;
}

void Wave::waveNum( unsigned int num)
{
	if ( ( num <= _waveCount ) && ( num > 0 ) )
		_waveNumber = num;
}

void Wave::FillHeader(short channels,long sampleRate,short dataSize, long totalSamples)
{
   Set(channels,sampleRate,dataSize);
   waveHeader.ch_data.chunksize=totalSamples * waveHeader.format.wf.nBlockAlign;
   waveHeader.header.filesize=waveHeader.ch_data.chunksize + 36;
}


Boolean Wave::LoadWav(String fileName)
{  
	if(wavFile)
	    fclose(wavFile);  

	if ((wavFile = fopen(fileName,"r+")) == NULL)
	    if ((wavFile = fopen(fileName,"r")) == NULL)
			return False;
  
	pName = new char[strlen(fileName) + 1];
	strcpy(pName, fileName);
 
	if(!wavFile)
   	Set(1,11025,8);	// minimalist defaults for a new file
	else
	{   
		fread(&waveHeader,sizeof(waveHeader),1,wavFile);

		if(!IsWavFile()) 
			return False; 
		else
		{
			if(waveType() != WAVE_FORMAT_PCM)
				return False;
			else
			{
				pData = new char[waveHeader.ch_data.chunksize];
				fread(pData,waveHeader.ch_data.chunksize,1,wavFile);   
			}
		}
	}
   
	return True;
}


Boolean Wave::LoadWav(String fileName,short wavType)
{ 
   if(wavFile)	fclose(wavFile);

   if ((wavFile = fopen(fileName,"r+")) == NULL)
	if ((wavFile = fopen(fileName,"r")) == NULL)
	    return False;
			    
   pName = new char[strlen(fileName) + 1];
   strcpy(pName, fileName);
 
   if(!wavFile)
      Set(1,11025,8);
   else
   {
      fread(&waveHeader,sizeof(waveHeader),1,wavFile);

      if(!IsWavFile())
        return False;
      else
      {
      	if(waveType() != WAVE_FORMAT_PCM)
            return False;
         else
         {
            pData = new char[waveHeader.ch_data.chunksize];                  
            fread(pData,waveHeader.ch_data.chunksize,1,wavFile);      
         }
      }
   }

   return True;
}


Boolean Wave::LoadHeader(String fileName)
{
    if(wavFile) fclose(wavFile);

    if ((wavFile = fopen(fileName,"r+")) == NULL)
        if ((wavFile = fopen(fileName,"r")) == NULL)
	    return False;
			    
	pName = new char[strlen(fileName) + 1];
	strcpy(pName, fileName);

   if(!wavFile)
      Set(1,11025,8);	// minimalist defaults for a new file
   else
   {
	  if (fread(&waveHeader,sizeof(waveHeader),1,wavFile) != 1 )
		 return False;

      if(!IsWavFile())
         return False;
         
      if(waveType() != WAVE_FORMAT_PCM )
         return False;
   }

   return True;
}


Boolean Wave::LoadHeader(String fileName,short wavType)
{
   if(wavFile) fclose(wavFile);
   
   if ((wavFile = fopen(fileName,"r+")) == NULL)
        if ((wavFile = fopen(fileName,"r")) == NULL)
	    return False;
			    
	pName = new char[strlen(fileName) + 1];
	strcpy(pName, fileName);

   if(!wavFile)
      Set(1,11025,8);   
   else
   {
      fread(&waveHeader,sizeof(waveHeader),1,wavFile);

      if(!IsWavFile())
         return False;

      if(waveType() != WAVE_FORMAT_PCM)
         return False;
   }
   
   return True;
}


Boolean Wave::SaveHeader(String fileName)
{
   FILE *outputWav;

   outputWav = fopen(fileName,"wb");

   if(!outputWav)
      return False;
   else
      fwrite(&waveHeader,sizeof(waveHeader),1,outputWav);

   fclose(outputWav);
   return True;
}


Boolean Wave::ChangeHeader(String fileName)
{
   FILE *outputWav;

   if ((outputWav = fopen(fileName,"r+")) == NULL)
        if ((outputWav = fopen(fileName,"r")) == NULL)
	    return False;
			    
   if(!outputWav)
      return False;
   else
      fwrite(&waveHeader,sizeof(waveHeader),1,outputWav);

   fclose(outputWav);
   return True;
}


Boolean Wave::SaveWav(String fileName)
{ 
  FILE *outputWav;
  outputWav = fopen(fileName, "wb");

  if(!outputWav)
     return False;
  else
  {
     fwrite(&waveHeader,sizeof(waveHeader),1,outputWav);
     fwrite(pData,waveHeader.ch_data.chunksize>>1,1,outputWav);
  }

  fclose(outputWav);
  return True;
}


long Wave::LoadBlock(long filePos, long blockLength)
{
   long localBlockSize;

   if(!wavFile) return 0;

	if (pData)
		delete(pData);  
   
	fseek(wavFile, filePos + sizeof(waveHeader), 0);

   if((waveHeader.ch_data.chunksize - filePos) > blockLength)
      localBlockSize = blockLength;
   else
   	localBlockSize = waveHeader.ch_data.chunksize - filePos;


   pData = new char[localBlockSize];
   return fread(pData,localBlockSize,1,wavFile);
}


Boolean Wave::SaveBlock(String fileName, long filePos, long blockLength, void *data )
{	// warning - only to be used for modifying data where header is unchanged
   if(!wavFile)
     return False;
   else
   {
	 fseek(wavFile, filePos+sizeof(waveHeader), SEEK_SET);
   	 fwrite(data, blockLength, 1, wavFile);
	 fclose ( wavFile );

	 if ((wavFile = fopen ( fileName, "r+")) == NULL)
	 	return False;
   }
   
   return True;
}


Boolean Wave::AppendBlock(String fileName,short channels,long sampleRate,
	short dataSize,long numSamples)
{  
   FILE *outputWav;
   FILE *checkWav;

   // see if a file of that name already exists
   checkWav = fopen(fileName,"rb");

   if(checkWav)
   {  
      // check that the output file is a .wav file
      if(!IsWavFile())
      {
      	fclose(checkWav);      
         return False;
      }
	  
	  fclose(checkWav);
   }
   
   // open output .wav file for appending 
   if ((outputWav = fopen(fileName,"a+")) == NULL)
	return False;
			       
   if(!outputWav)
      return False;

   // write the extra wave data
   if(!pData)
   {
      printf ( "waveclass error: append block called with no data!\n" );
		// flag this otherwise untraceable error!
      return False;							   
   }
   else
   {
      fwrite(pData,numSamples*waveHeader.format.wf.nBlockAlign,1,outputWav);
   }   

   fclose(outputWav);
   return True;
}


void Wave::RewindToStart(void)
{
   if(wavFile)
      rewind ( wavFile );
}


Boolean Wave::MajMin(String fileName)
{  // big/little endian HEADER ONLY interconversion
   FILE *outputWav;
   FILE *checkWav; 
   char arrayI[44];
   char arrayO[44];
   int i;

   if(wavFile)	fclose(wavFile);   
   checkWav = fopen(fileName,"r");

   if(checkWav)
   {  // check that the input file is a .wav file of some sort at least
      if(!(IsWavFile() || IsAltWavFile()))
      {
      	fclose(checkWav);      
         return False;
      }		
   }
    
   fread(&arrayI,44,1,wavFile);
   fclose(checkWav);
   
   for (i = 0; i < 4; i++)
      arrayO[4 - i] = arrayI[i];	// 'RIFF' / 'FFIR', etc...
      
   arrayO[4] = arrayI[6];
   arrayO[5] = arrayI[7];
   arrayO[6] = arrayI[4];
   arrayO[7] = arrayI[5];
  
   for (i = 0; i < 4; i++)
      arrayO[11 - i] = arrayI[i + 8];
      
   for (i = 0; i < 4; i++)
      arrayO[15 - i] = arrayI[i + 12];
 
   arrayO[16] = arrayI[18];
   arrayO[17] = arrayI[19];
   arrayO[18] = arrayI[16];
   arrayO[19] = arrayI[17];
   
   arrayO[20] = arrayI[21];
   arrayO[21] = arrayI[20];
   
   arrayO[22] = arrayI[23];
   arrayO[23] = arrayI[22];

   for (i = 0; i < 4; i++)
      arrayO[27 - i] = arrayI[i + 24];
     
   arrayO[28] = arrayI[30];
   arrayO[29] = arrayI[31];
   arrayO[30] = arrayI[28];
   arrayO[31] = arrayI[29];
  
   arrayO[32] = arrayI[33];
   arrayO[33] = arrayI[32];
   
   arrayO[34] = arrayI[35];
   arrayO[35] = arrayI[34];
   
   for (i = 0; i < 4; i++)
      arrayO[39 - i] = arrayI[i + 36];
   
   for (i = 0; i < 4; i++)
      arrayO[43 - i] = arrayI[i + 40];

   outputWav = fopen(fileName,"r+");

   if(!outputWav)
      return False;
   else
   {
      fseek(outputWav, 0L, SEEK_SET);   
      fwrite(&arrayO,44,1,outputWav);
      fclose(outputWav);      
      return True;
   }
}   


char *Wave::name(void) {return pName;}
short Wave::waveType(void) { return waveHeader.format.wf.wFormatTag; }
short Wave::numChannels(void) { return waveHeader.format.wf.nChannels; }
short Wave::bitsPerSample(void) { return waveHeader.format.wBitsPerSample; }
long  Wave::sampleRate(void) { return waveHeader.format.wf.nSamplesPerSec; }
short Wave::blockAlign(void) { return waveHeader.format.wf.nBlockAlign; }
long  Wave::dataLength(void) { return waveHeader.ch_data.chunksize; }
long  Wave::fileSize(void) { return waveHeader.header.filesize; }
char *Wave::riffHeader(void) { return waveHeader.header.riff; }

char *Wave::dataChunkId(void)
{
	return waveHeader.ch_data.chunk_id;
}

long  Wave::formatLength(void) { return waveHeader.ch_format.chunksize; }

char *Wave::formatChunkId(void)
{
	return ( char * ) waveHeader.ch_format.chunk_id;
}
/*

char *Wave::riffHeader(void)
{
	return ( char * ) waveHeader.header.riff;
}
*/

char *Wave::riffType(void)
{
	return ( char * ) waveHeader.header.rifftype;
}



long Wave::totalSamples(void)
{
   return waveHeader.ch_data.chunksize / waveHeader.format.wf.nBlockAlign;
}

 
WaveFile *Wave::GetHeader(void)
{
   return &waveHeader;
}


char *Wave::data(void)
{
   return pData;
}


// only needed for writing .wav files
Boolean Wave::AllocateDataSpace(long bytes)
{
   pData = new char[bytes];
   
   if(pData)
      return True;
   else
      return False;
}

#endif
